/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#pragma once
#include <ders/config.hpp>
#include <exception>
#include <system_error>
#include <ders/tx_buf.hpp>

#define FLLN ders::FileLine(__FILE__, __LINE__)

namespace ders {

struct FileLine {
    const char* const file;
    const int line;

    FileLine(const char* f, int l) : file(f), line(l) {}
};

class Exception : public deletable {
 public:
    Exception(mem_pool* mp, FileLine loc, strsz msg) : loc_(loc), msg_(mp, msg), nest_(mp) {}
    Exception(FileLine loc, strsz msg, sh_ptr<Exception> nest) : loc_(loc), msg_(ders::pool(nest), msg), nest_(nest) {}

    mem_pool* pool() const { return msg_.pool(); }
    FileLine location() const { return loc_; }
    strsz message() const { return msg_.str(); }
    sh_ptr<Exception> nested() const { return nest_; }

    virtual strsz className() const { return strsz("ders::Exception"); }
    virtual un_text toText() const;

    template<class E> E* is() { return dynamic_cast<E*>(this); }
    template<class E> const E* is() const { return dynamic_cast<E*>(this); }

    Exception(const Exception&)=delete;
    Exception& operator=(const Exception&)=delete;

    void delete_this(mem_pool* mp) const override { delete_ptr(mp, this); }

 protected:
    FileLine loc_;
    text msg_;
    sh_ptr<Exception> nest_;
};

class ExternalException : public Exception {
 public:
    ExternalException(mem_pool* mp, FileLine loc, strsz msg, std::exception_ptr eptr) : Exception(mp, loc, msg), eptr_(eptr) {}

    std::exception_ptr external() const { return eptr_; }

    strsz className() const override { return strsz("ders::ExternalException"); }
    void delete_this(mem_pool* mp) const override { delete_ptr(mp, this); }

 protected:
    std::exception_ptr eptr_;
};

class StdException : public ExternalException {
 public:
    StdException(mem_pool* mp, FileLine loc, strsz msg, std::exception_ptr eptr, const std::type_info* ti) :
        ExternalException(mp, loc, msg, eptr), ti_(ti) {}

    const std::type_info* type() const { return ti_; }

    strsz className() const override { return strsz("ders::StdException"); }
    un_text toText() const override;
    void delete_this(mem_pool* mp) const override { delete_ptr(mp, this); }

 protected:
    const std::type_info* ti_;
};

class ErrorException : public Exception {
 public:
    ErrorException(mem_pool* mp, FileLine loc, strsz msg, std::error_code ec) : Exception(mp, loc, msg), ec_(ec) {}
    ErrorException(FileLine loc, strsz msg, std::error_code ec, sh_ptr<Exception> nest) : Exception(loc, msg, nest), ec_(ec) {}

    std::error_code error() const { return ec_; }

    strsz className() const override { return strsz("ders::ErrorException"); }
    un_text toText() const override;
    void delete_this(mem_pool* mp) const override { delete_ptr(mp, this); }

 protected:
    std::error_code ec_;
};

class FileException : public ErrorException {
 public:
    FileException(mem_pool* mp, FileLine loc, strsz msg, std::error_code ec, strsz path) :
        ErrorException(mp, loc, msg, ec), path_(mp, path) {}
    FileException(FileLine loc, strsz msg, std::error_code ec, strsz path, sh_ptr<Exception> nest) :
        ErrorException(loc, msg, ec, nest), path_(ders::pool(nest), path) {}

    strsz path() const { return path_.str(); }

    strsz className() const override { return strsz("ders::FileException"); }
    un_text toText() const override;
    void delete_this(mem_pool* mp) const override { delete_ptr(mp, this); }

 protected:
    text path_;
};

class MsgException : public Exception {
 public:
    MsgException(mem_pool* mp, FileLine loc, strsz msg) : Exception(mp, loc, msg) {}

    strsz className() const=0;
    void delete_this(mem_pool* mp) const=0;
};

class ExitMsgException : public MsgException {
 public:
    ExitMsgException(mem_pool* mp, FileLine loc, strsz msg, int ecode) : MsgException(mp, loc, msg), ecd_(ecode) {}

    int exitCode() const { return ecd_; }

    strsz className() const override { return strsz("ders::ExitMsgException"); }
    un_text toText() const override;
    void delete_this(mem_pool* mp) const override { delete_ptr(mp, this); }

 protected:
    int ecd_;
};

template<class E, class... Args>
inline sh_ptr<Exception> newExc(mem_pool* mp, Args&&... args)
{
    return sh_ptr<Exception>(make_un<E>(mp, std::forward<Args>(args)...));
}

un_text toTextAll(const Exception* exc);
sh_ptr<Exception> recatch(mem_pool* mp, FileLine loc, bool rethrowMsg=true);

}  // namespace ders
