/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#pragma once
#include <ders/config.hpp>
#include <ders/exception.hpp>
#include <ders/rdwr.hpp>

namespace ders {

const int fd_in= 0;
const int fd_out=1;
const int fd_err=2;

const int md_rdo =0b00000001;
const int md_wro =0b00000010;
const int md_rdwr=0b00000100;
const int mask_md=0b00000111;

const int fl_app =0b00001000;
const int fl_crt =0b00010000;
const int fl_excl=0b00100000;
const int fl_trnc=0b01000000;
const int mask_fl=0b01111000;

enum whence {
    wh_set=1,
    wh_cur=2,
    wh_end=3
};

namespace impl {

int fd_open(const char* path, int flags, std::error_code* ec);    
bool fd_close(int fd, std::error_code* ec);

int fd_read(int fd, void* buf, int n, std::error_code* ec);
int fd_write(int fd, const void* buf, int n, std::error_code* ec);

long long fd_seek(int fd, long long off, whence whn, std::error_code* ec);

}  // namespace impl

class fd_file : public readable, public writable {
 public:
    fd_file(mem_pool* mp, strsz path, int flags, std::error_code* ec) : path_(mp, path), own_(true)
    {
        fd_=impl::fd_open(path_.c_str(), flags, ec);
    }
    fd_file(mem_pool* mp, strsz path, int flags);

    fd_file(mem_pool* mp, int fd) : path_(mp, tx_buf(mp)+'<'+fd+'>'), fd_(fd), own_(false) { assert(fd>=0); }

    ~fd_file() 
    { 
        std::error_code ignore;
        if (own_ && fd_>=0) impl::fd_close(fd_, &ignore); 
    }

    explicit operator bool() const { return fd_>=0; }
    
    mem_pool* pool() const { return path_.pool(); }
    strsz path() const { return path_.str(); }

    int read(void* buf, int n, std::error_code* ec) override { return impl::fd_read(fd_, buf, n, ec); }
    int ex_read(void* buf, int n);

    int write(const void* buf, int n, std::error_code* ec) override { return impl::fd_write(fd_, buf, n, ec); }
    int ex_write(const void* buf, int n);

    int write(strsz s, std::error_code* ec) { return write(s.str(), s.size(), ec); }
    int ex_write(strsz s) { return ex_write(s.str(), s.size()); }

    long long tell(std::error_code* ec) { return seek(0, wh_cur, ec); }
    long long ex_tell();

    long long seek(long long off, whence whn, std::error_code* ec) { return impl::fd_seek(fd_, off, whn, ec); }
    long long ex_seek(long long off, whence whn);

    fd_file(const fd_file&)=delete;
    fd_file& operator=(const fd_file&)=delete;

 private:    
    text path_;
    int fd_;
    bool own_;
};

}  // namespace ders
