/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#pragma once
#include <ders/config.hpp>
#include <memory>
#include <ders/mem_pool.hpp>

namespace ders {

template<class T>
class mp_allocator {
 public:
    using value_type=T;
    using is_always_equal=std::false_type;
    using propagate_on_container_copy_assignment=std::true_type;
    using propagate_on_container_move_assignment=std::true_type;
    using propagate_on_container_swap=std::true_type;

    explicit mp_allocator(mem_pool* mp) : mp_(mp) {}
    template<class U> mp_allocator(const mp_allocator<U>& ma) : mp_(ma.pool()) {}

    T* allocate(size_t n) { return n ? static_cast<T*>(mp_->alloc(n*sizeof(T))) : 0; }
    void deallocate(T* p, size_t n) { if (n) mp_->free(p, n*sizeof(T)); }

    mem_pool* pool() const { return mp_; }

#ifdef DERS_OLD_ALLOCATOR
    template<class U> struct rebind { typedef mp_allocator<U> other; };

    using size_type=size_t;
    using difference_type=ptrdiff_t;
    using pointer=T*;
    using const_pointer=const T*;
    using reference=T&;
    using const_reference=const T&;

    pointer address(reference x) const { return &x; }
    const_pointer address(const_reference x) const { return &x; }
    size_type max_size() const { return size_t(-1)/sizeof(T); }

    template<class Up, class... Args>
    void construct(Up* p, Args&&... args) { ::new((void*)p) Up(std::forward<Args>(args)...); }

    template<class Up> void destroy(Up* p) { p->~Up(); }
#endif

 private:
    mem_pool* mp_;
};

template<class T1, class T2>
bool operator==(const mp_allocator<T1>& ma1, const mp_allocator<T2>& ma2) { return ma1.pool()==ma2.pool(); }

template<class T1, class T2>
bool operator!=(const mp_allocator<T1>& ma1, const mp_allocator<T2>& ma2) { return !(ma1==ma2); }

}  // namespace ders
