/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#pragma once
#include <ders/config.hpp>
#include <new>
#include <typeinfo>
#include <utility>
#include <assert.h>
#include <ders/mem_pool.hpp>

namespace ders {

class deletable {
 public:
    virtual void delete_this(mem_pool* mp) const=0;
    virtual ~deletable() {}

 protected:
    template<class T> static void delete_ptr(mem_pool* mp, T* ptr);
};

namespace impl {

template<class T>
inline void mp_del(mem_pool* mp, const void* ptr)
{
    auto tptr=(T*)ptr;

    assert(typeid(*tptr)==typeid(T));
    tptr->~T();

    mp->free((void*)tptr, sizeof(T));
}

template<class T>
inline void mp_del(mem_pool* mp, const deletable* ptr)
{
    ptr->delete_this(mp);
}

}  // namespace impl

template<class T>
inline void deletable::delete_ptr(mem_pool* mp, T* ptr)
{
    impl::mp_del<T>(mp, (const void*)ptr);
}

template<class T, class... Args>
inline T* mp_new(mem_pool* mp, Args&&... args)
{
    mp_buf b(mp, sizeof(T));
    new(b.buf()) T(std::forward<Args>(args)...);

    return static_cast<T*>(b.release());
}

template<class T>
inline void mp_delete(mem_pool* mp, T* ptr)
{
    impl::mp_del<T>(mp, ptr);
}

}  // namespace ders
