/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#pragma once
#include <ders/config.hpp>
#include <type_traits>
#include <ders/mp_new.hpp>

namespace ders {

template<class T>
class un_ptr {
    template<class> friend class un_ptr;
    using RCT=std::remove_const_t<T>;

 public:
    explicit un_ptr(mem_pool* mp, T* ptr=0) : mp_(mp), ptr_(ptr) { assert(mp); }

    un_ptr(const un_ptr&)=delete;
    un_ptr(un_ptr&& un) noexcept : mp_(un.mp_), ptr_(un.ptr_) { un.ptr_=0; }

    // construct un_ptr<const Type> from un_ptr<Type>
    template<class T2, class=std::enable_if_t<!std::is_same<T, RCT>::value && std::is_same<T2, RCT>::value>>
    un_ptr(un_ptr<T2>&& un) noexcept : mp_(un.mp_), ptr_(un.ptr_) { un.ptr_=0; }

    ~un_ptr() { delete_ptr(); }

    un_ptr& operator=(const un_ptr&)=delete;
    un_ptr& operator=(un_ptr&& un) noexcept
    {
        if (this!=&un) {
            delete_ptr();

            mp_=un.mp_;
            ptr_=un.ptr_;

            un.ptr_=0;
        }

        return *this;
    }

    T& operator*() const { return *ptr_; }
    T* operator->() const { return ptr_; }
    explicit operator bool() const { return ptr_; }

    template<class T2> friend mem_pool* pool(const un_ptr<T2>& self);
    template<class T2> friend T2* get(const un_ptr<T2>& self);
    template<class T2> friend T2* release(un_ptr<T2>& self);
    template<class T2> friend void swap(un_ptr<T2>& un1, un_ptr<T2>& un2) noexcept;

 private:
    mem_pool* mp_;
    T* ptr_;

    void delete_ptr() { if (ptr_) mp_delete(mp_, ptr_); }
};

template<class T2>
inline mem_pool* pool(const un_ptr<T2>& self) { return self.mp_; }

template<class T2>
inline T2* get(const un_ptr<T2>& self) { return self.ptr_; }

template<class T2>
inline T2* release(un_ptr<T2>& self)
{
    T2* ret=self.ptr_;
    self.ptr_=0;
    return ret;
}

template<class T2>
inline void swap(un_ptr<T2>& un1, un_ptr<T2>& un2) noexcept
{
    std::swap(un1.mp_, un2.mp_);
    std::swap(un1.ptr_, un2.ptr_);
}

template<class T, class... Args>
inline un_ptr<T> make_un(mem_pool* mp, Args&&... args)
{
    return un_ptr<T>(mp, mp_new<T>(mp, std::forward<Args>(args)...));
}

template<class T>
class sh_ptr {
    template<class> friend class sh_ptr;
    using RCT=std::remove_const_t<T>;

    struct data {
        mem_pool* mp_;
        T* ptr_;
        int cnt_;

        data(mem_pool* mp, T* ptr) : mp_(mp), ptr_(ptr), cnt_(1) {}
        ~data() { delete_ptr(); }

        void delete_ptr() { if (ptr_) mp_delete(mp_, ptr_); }
    };

 public:
    explicit sh_ptr(mem_pool* mp, T* ptr=0) : dt_(mp_new<data>(mp, mp, ptr)) {}

    template<class T2>
    explicit sh_ptr(un_ptr<T2>&& un) : dt_(mp_new<data>(pool(un), pool(un), get(un))) { release(un); }

    sh_ptr(const sh_ptr& sh) noexcept : dt_(sh.dt_) { dt_->cnt_++; }

    // construct sh_ptr<const Type> from sh_ptr<Type>
    template<class T2, class=std::enable_if_t<!std::is_same<T, RCT>::value && std::is_same<T2, RCT>::value>>
    sh_ptr(const sh_ptr<T2>& sh) noexcept : dt_(reinterpret_cast<decltype(dt_)>(sh.dt_)) { dt_->cnt_++; }

    ~sh_ptr() { dec_delete(); }

    sh_ptr& operator=(const sh_ptr& sh) noexcept
    {
        sh.dt_->cnt_++;  // self assignment
        dec_delete();
        dt_=sh.dt_;

        return *this;
    }

    template<class T2>
    sh_ptr& operator=(un_ptr<T2>&& un)
    {
        auto mp2=pool(un);
        if (dt_->cnt_==1 && dt_->mp_==mp2) {
            dt_->delete_ptr();
            dt_->ptr_=release(un);
        }
        else {
            auto ndt=mp_new<data>(mp2, mp2, get(un));
            release(un);

            dec_delete();
            dt_=ndt;
        }

        return *this;
    }

    T& operator*() const { return *dt_->ptr_; }
    T* operator->() const { return dt_->ptr_; }
    explicit operator bool() const { return dt_->ptr_; }

    template<class T2> friend mem_pool* pool(const sh_ptr<T2>& self);
    template<class T2> friend T2* get(const sh_ptr<T2>& self);
    template<class T2> friend int count(const sh_ptr<T2>& self);
    template<class T2> friend T2* release(sh_ptr<T2>& self);
    template<class T2> friend void swap(sh_ptr<T2>& sh1, sh_ptr<T2>& sh2) noexcept;

 private:
    data* dt_;

    void dec_delete() { if (--dt_->cnt_==0) mp_delete(dt_->mp_, dt_); }
};

template<class T2>
inline mem_pool* pool(const sh_ptr<T2>& self) { return self.dt_->mp_; }

template<class T2>
inline T2* get(const sh_ptr<T2>& self) { return self.dt_->ptr_; }

template<class T2>
inline int count(const sh_ptr<T2>& self) { return self.dt_->cnt_; }

template<class T2>
inline T2* release(sh_ptr<T2>& self)
{
    T2* ret=self.dt_->ptr_;
    self.dt_->ptr_=0;
    return ret;
}

template<class T2>
inline void swap(sh_ptr<T2>& sh1, sh_ptr<T2>& sh2) noexcept
{
    std::swap(sh1.dt_, sh2.dt_);
}

template<class T, class... Args>
inline sh_ptr<T> make_sh(mem_pool* mp, Args&&... args)
{
    return sh_ptr<T>(make_un<T>(mp, std::forward<Args>(args)...));
}

}  // namespace ders
