/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#pragma once
#include <ders/config.hpp>
#include <system_error>
#include <ders/text.hpp>

namespace ders {

class readable {
 public:
    virtual int read(void* buf, int n, std::error_code* ec)=0;
    virtual ~readable() {}
};

class writable {
 public:
    virtual int write(const void* buf, int n, std::error_code* ec)=0;
    virtual ~writable() {}
};

class buf_rd {
 public:
    buf_rd(mem_pool* mp, readable* rd, int bufsz=16*1024) : rd_(rd), mb_(mp, bufsz) {}

    mem_pool* pool() const { return mb_.pool(); }

    explicit operator bool() const { return !ec_; }
    std::error_code error() const { return ec_; }

    int get() { return (cur_<end_) ? beg()[cur_++] : get_impl(); }
    bool get_line(text* line, bool noendl=true);

    int read(void* buf, int n);
    int read(text* t, int n);

    buf_rd(const buf_rd&)=delete;
    buf_rd& operator=(const buf_rd&)=delete;

 private:    
    readable* rd_;
    std::error_code ec_;

    mp_buf mb_;
    int cur_=0;
    int end_=0;

    unsigned char* beg() const { return static_cast<unsigned char*>(mb_.buf()); }
    int read_next();
    int get_impl() { return (read_next()>0) ? beg()[cur_++] : -1; }
};

class buf_wr {
 public:
    buf_wr(mem_pool* mp, writable* wr, int bufsz=16*1024) : wr_(wr), mb_(mp, bufsz) {}
    ~buf_wr() { flush(); }

    mem_pool* pool() const { return mb_.pool(); }

    explicit operator bool() const { return !ec_; }
    std::error_code error() const { return ec_; }

    int put(int ch)  { return (cur_<mb_.size()) ? beg()[cur_++]=ch : put_impl(ch); }

    int write(const void* buf, int n);
    int write(strsz s) { return write(s.str(), s.size()); }

    bool flush() { write_next(); return cur_==0; }

    buf_wr(const buf_wr&)=delete;
    buf_wr& operator=(const buf_wr&)=delete;

 private:    
    writable* wr_;
    std::error_code ec_;

    mp_buf mb_;
    int cur_=0;

    unsigned char* beg() const { return static_cast<unsigned char*>(mb_.buf()); }
    int write_next();
    unsigned char put_impl(int ch) { return (write_next()>0) ? beg()[cur_++]=ch : -1; }
};

}  // namespace ders
