/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#pragma once
#include <ders/config.hpp>
#include <string.h>
#include <ders/ptr.hpp>

namespace ders {

namespace impl {

template<class T>
struct wrap {
    T v_;
    wrap(T v) : v_(v) {}
};

}  // namespace impl

class strsz {
 public:
    strsz() : str_(0), sz_(0) { invariant(); }
    strsz(const char* str, int sz) : str_(str), sz_(sz) { invariant(); }
    strsz(const char* beg, const char* end) : str_(beg), sz_(end-beg) { invariant(); }

    template<size_t N> strsz(const char (&chn)[N]) : str_(chn), sz_(N-1) { invariant(); }
    strsz(impl::wrap<const char*> w) : str_(w.v_), sz_(strlen(w.v_)) { invariant(); }

    char operator[](int i) const { assert(i>=0 && i<sz_); return str_[i]; }

    const char* str() const { return str_; }
    int size() const { return sz_; }

    const char* begin() const { return str_; }
    const char* end() const { return str_+sz_; }

    char front() const { assert(sz_>0); return str_[0]; }
    char back() const { assert(sz_>0);  return str_[sz_-1]; }

    bool begins(strsz s) const { return sz_>=s.sz_ && memcmp(str_, s.str_, s.sz_)==0; }
    bool ends(strsz s) const { return sz_>=s.sz_ && memcmp(str_+sz_-s.sz_, s.str_, s.sz_)==0; }

    const char* find(char ch) const { return (const char*)memchr(str_, ch, sz_); }
    const char* find(strsz s) const;

 private:
    const char* str_;
    int sz_;

    void invariant() { assert((str_ && sz_>=0) || (!str_ && !sz_)); }
};

namespace impl {

inline int strszcmp(strsz s1, strsz s2)
{
    int sz1=s1.size(), sz2=s2.size();
    int diff=memcmp(s1.str(), s2.str(), (sz1<sz2) ? sz1 : sz2);
    return (diff) ? diff : sz1-sz2;
}

}  // namespace impl

inline bool operator==(strsz s1, strsz s2) { return s1.size()==s2.size() && memcmp(s1.str(), s2.str(), s1.size())==0; }
inline bool operator!=(strsz s1, strsz s2) { return !(s1==s2); }
inline bool operator< (strsz s1, strsz s2) { return impl::strszcmp(s1, s2)<0; }
inline bool operator<=(strsz s1, strsz s2) { return impl::strszcmp(s1, s2)<=0; }
inline bool operator> (strsz s1, strsz s2) { return impl::strszcmp(s1, s2)>0; }
inline bool operator>=(strsz s1, strsz s2) { return impl::strszcmp(s1, s2)>=0; }

class text {
 public:
    text(mem_pool* mp, strsz s, int rsrv=0) { init(mp, s, rsrv+1); }
    explicit text(mem_pool* mp, int rsrv=0) { init(mp, {}, rsrv+1); }
    text(const text&)=delete;
    ~text() { mp_->free(buf_, resd_); }

    text& operator=(const text&)=delete;
    text& operator=(strsz s);

    text& operator+=(strsz s);
    text& operator+=(char ch)
    {
        reserve(used_+1);
        buf_[used_++]=ch;

        return *this;
    }

    char& operator[](int i) { assert(i>=0 && i<used_); return buf_[i]; }
    char operator[](int i) const { assert(i>=0 && i<used_); return buf_[i]; }

    mem_pool* pool() const { return mp_; }
    strsz str() const { return strsz(buf_, used_); }
    int size() const { return used_; }
    int capacity() const { return resd_-1; }

    const char* c_str() const { buf_[used_]=0; return buf_; }

    void clear() { used_=0; }
    void reserve(int sz) { if (sz+1>resd_) reserve_impl(sz+1); }
    void uninitialized_resize(int sz) { reserve(sz); used_=sz; }

    char* begin() { return buf_; }
    const char* begin() const { return buf_; }
    const char* cbegin() const { return buf_; }

    char* end() { return buf_+used_; }
    const char* end() const { return buf_+used_; }
    const char* cend() const { return buf_+used_; }

    char& front() { return *begin(); }
    char front() const { return *begin(); }

    char& back() { return *(end()-1); }
    char back() const { return *(end()-1); }

 private:
    mem_pool* mp_;
    char* buf_;
    int resd_;
    int used_;

    void init(mem_pool* mp, strsz s, int real);
    void reserve_impl(int real);
};

using un_text=un_ptr<text>;
inline un_text untx(mem_pool* mp, strsz s) { return make_un<text>(mp, mp, s); }

using sh_text=sh_ptr<text>;
inline sh_text shtx(mem_pool* mp, strsz s) { return make_sh<text>(mp, mp, s); }

inline bool operator==(const sh_text& st1, const sh_text& st2) { return st1->str()==st2->str(); }
inline bool operator!=(const sh_text& st1, const sh_text& st2) { return st1->str()!=st2->str(); }
inline bool operator< (const sh_text& st1, const sh_text& st2) { return st1->str()<st2->str(); }
inline bool operator<=(const sh_text& st1, const sh_text& st2) { return st1->str()<=st2->str(); }
inline bool operator> (const sh_text& st1, const sh_text& st2) { return st1->str()>st2->str(); }
inline bool operator>=(const sh_text& st1, const sh_text& st2) { return st1->str()>=st2->str(); }

}  // namespace ders
