/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#include <ders/mem_pool.hpp>
#ifndef NDEBUG
#include <unordered_map>
#endif
#include <vector>
#include <assert.h>
#include <stdio.h>

namespace {

using namespace std;
using namespace ders;

const size_t SZ_LIM_1=512;
const size_t SZ_LIM_2=SZ_LIM_1+1024;
const size_t SZ_CHUNK=64*1024;

const size_t SZ_PTR=sizeof(void*);
const size_t SZ_PTR2=SZ_PTR*2;  // the alignment

const size_t CNT_HEADS1=SZ_LIM_1/SZ_PTR;
const size_t CNT_HEADS2=(SZ_LIM_2-SZ_LIM_1)/SZ_PTR2;

inline void* glob_alloc(size_t sz)
{
    return operator new(sz);  // can throw
}

inline void glob_free(void* ptr)
{
    operator delete(ptr);
}

inline size_t alignSize(size_t sz, size_t to)
{
    return (sz+to-1)/to*to;
}

}  // unnamed

namespace ders {

namespace impl {

struct mem_pool_data {
    void* heads1[CNT_HEADS1]{};
    void* heads2[CNT_HEADS2]{};

    vector<void*> chunks;
    char* chunk_beg=0;
    char* chunk_end=0;

#ifndef NDEBUG
    unordered_map<void*, size_t> allocated;
#endif

    ~mem_pool_data()
    {
        for (auto it=chunks.begin(), end=chunks.end(); it!=end; ++it) glob_free(*it);
    }

    void** headAddr(size_t sz)
    {
        assert(sz>0 && sz<=SZ_LIM_2);
        return (sz<=SZ_LIM_1) ? &heads1[(sz-1)/SZ_PTR] : &heads2[(sz-SZ_LIM_1-1)/SZ_PTR2];
    }

    void new_chunk();
    void* alloc(size_t sz);
};

void mem_pool_data::new_chunk()
{
    size_t sz=SZ_CHUNK-SZ_PTR2;  // SZ_PTR2 is glob_alloc() overhead

    chunks.reserve(chunks.size()+1);
    chunks.push_back(glob_alloc(sz));

    chunk_beg=(char*)chunks.back();
    chunk_end=chunk_beg+sz;
}

void* mem_pool_data::alloc(size_t sz)
{
    assert(sz>0 && sz<=SZ_LIM_2);

    size_t szf=chunk_end-chunk_beg, sza=alignSize(sz, (sz<=SZ_LIM_1) ? SZ_PTR : SZ_PTR2);
    if (szf<sza) {
        if (szf>0) {  // "free" it
            void** phead=headAddr(szf);

            *reinterpret_cast<void**>(chunk_beg)=*phead;
            *phead=chunk_beg;
        }

        new_chunk();
        assert(size_t(chunk_end-chunk_beg)>=sza);
    }

    void* ret;

    if (sza%SZ_PTR2) {
        chunk_end-=sza;
        ret=chunk_end;
    }
    else {
        ret=chunk_beg;
        chunk_beg+=sza;
    }

    return ret;
}

}  // namespace impl

mem_pool::mem_pool() : data(new impl::mem_pool_data)
{
}

mem_pool::~mem_pool()
{
#ifndef NDEBUG
    if (int sz=data->allocated.size()) {
        fprintf(stderr, "\n~mem_pool(%p): data->allocated.size()=%d\n", (void*)this, sz);

        int i=0;
        for (auto it=data->allocated.begin(), end=data->allocated.end(); it!=end; ++it) {
            fprintf(stderr, "(%p,%d)%c", it->first, int(it->second), (++i<sz) ? ' ' : '\n');
        }
    }

    assert(data->allocated.empty());
#endif

    delete data;
}

void* mem_pool::alloc(size_t sz)
{
    assert(sz>0);
    void* ret;

    if (sz<=SZ_LIM_2) {
        void** phead=data->headAddr(sz);

        if (*phead) {
            ret=*phead;
            *phead=*reinterpret_cast<void**>(*phead);
        }
        else ret=data->alloc(sz);
    }
    else ret=glob_alloc(sz);

#ifndef NDEBUG
    auto emplaced=data->allocated.emplace(ret, sz);
    assert(emplaced.second);
#endif

    return ret;
}

void mem_pool::free(void *ptr, size_t sz)
{
    assert(ptr && sz>0);

#ifndef NDEBUG
    auto found=data->allocated.find(ptr);
    assert(found!=data->allocated.end());
    assert(found->second==sz);
    data->allocated.erase(found);
#endif

    if (sz<=SZ_LIM_2) {
        void** phead=data->headAddr(sz);

        *reinterpret_cast<void**>(ptr)=*phead;
        *phead=ptr;
    }
    else glob_free(ptr);
}

}  // namespace ders
