/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#include <ders/fd_file.hpp>
#include <fcntl.h>
#include <unistd.h>

namespace {

using namespace std;    

template<class T>
inline T set_ec(T rc, error_code* ec)
{
    if (rc==-1) ec->assign(errno, system_category());
    return rc;
}

}  // unnamed

namespace ders {

using namespace std;    

namespace impl {

int fd_open(const char* path, int flags, error_code* ec)
{
    ec->clear();
    int f=0;

    switch (flags&mask_md) {
    case md_rdo:
        f=O_RDONLY;
        break;
    case md_wro:
        f=O_WRONLY;
        break;
    case md_rdwr:
        f=O_RDWR;
        break;
    default:
        *ec=make_error_code(errc::invalid_argument);
        return -1;
    }

    if (flags&fl_app)  f|=O_APPEND;
    if (flags&fl_crt)  f|=O_CREAT;
    if (flags&fl_excl) f|=O_EXCL;
    if (flags&fl_trnc) f|=O_TRUNC;

#ifdef O_BINARY
    f|=O_BINARY;
#endif    

    return set_ec(open(path, f, 0644), ec);
}

bool fd_close(int fd, error_code* ec)
{
    assert(fd>=0);
    ec->clear();

    return set_ec(close(fd), ec)==0;
}

int fd_read(int fd, void* buf, int n, error_code* ec)
{
    assert(fd>=0 && n>0);
    ec->clear();

    int rc=read(fd, buf, n);
    assert(rc<=n);

    return set_ec(rc, ec);
}

int fd_write(int fd, const void* buf, int n, error_code* ec)
{
    assert(fd>=0 && n>0);
    ec->clear();

    int rc=write(fd, buf, n);
    assert(rc<=n);

    return set_ec(rc, ec);
}

long long fd_seek(int fd, long long off, whence whn, error_code* ec)
{
    assert(fd>=0);
    ec->clear();

    int w=0;

    switch (whn) {
    case wh_set:
        w=SEEK_SET;
        break;
    case wh_cur:
        w=SEEK_CUR;
        break;
    case wh_end:
        w=SEEK_END;
        break;
    default:
        *ec=make_error_code(errc::invalid_argument);
        return -1;
    }

    static_assert(sizeof(decltype(lseek(0, 0, 0)))>4, "long long required");
    return set_ec(lseek(fd, off, w), ec);
}

}  // namespace impl

}  // namespace ders
