/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#include <ders/latch.hpp>
#include <condition_variable>

namespace {

using namespace std;
using namespace ders;

struct latch_impl : public latch {
    int cnt_;
    void (*fun_)(void*);
    void* arg_;

    mutex mx_;
    condition_variable cv_;
    int wcnt_=0;
    bool interrupted_=false;

    latch_impl(int cnt, void (*on_start)(void*), void* arg);

    bool arrive_and_wait() override;
    void interrupt() override;

    void delete_this(mem_pool* mp) const override { delete_ptr(mp, this); }
};

latch_impl::latch_impl(int cnt, void (*on_start)(void*), void* arg) : cnt_(cnt), fun_(on_start), arg_(arg)
{
    assert(cnt>0);
}

bool latch_impl::arrive_and_wait()
{
    unique_lock<mutex> lock(mx_);
    if (interrupted_) return false;    

    if (++wcnt_==cnt_) {
        if (fun_) fun_(arg_);
        lock.unlock();

        cv_.notify_all();
        return true;
    }

    for (;;) {
        if (interrupted_) return false;    
        if (wcnt_>=cnt_) return true;

        cv_.wait(lock);
    }
}

void latch_impl::interrupt()
{
    {
        lock_guard<mutex> lg(mx_);
        interrupted_=true;
    }
    
    cv_.notify_all();
}

}  // unnamed

namespace ders {

un_ptr<latch> new_latch(mem_pool* mp, int cnt, void (*on_start)(void*), void* arg)    
{
    return un_ptr<latch>(mp, mp_new<latch_impl>(mp, cnt, on_start, arg));
}

}  // namespace ders
