/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#include "incfile.hpp"
#include <ders/dir.hpp>
#include <ders/fd_file.hpp>
#include <ders/hash.hpp>

namespace {

using namespace std;
using namespace ders;
using namespace inchk;

struct IncFileDBImpl : public IncFileDB {
    using UMapInt=mp_unordered_map<sh_text, int>;
    using AllocInt=UMapInt::allocator_type;

    using UMapData=mp_unordered_map<sh_text, sh_ptr<IncFileData>>;
    using AllocData=UMapData::allocator_type;

    mem_pool* mp_;
    vector<IncFilePath> vifp_;
    UMapInt keyInc_;
    UMapInt keyFull_;
    UMapData tblData_;

    IncFileDBImpl(mem_pool* mp, const vector<sh_text>* incDirs);

    const IncFilePath* findByIncName(strsz in) const override;
    const IncFilePath* findByFullPath(strsz fp) const override;
    const IncFileData* getData(strsz path) override;

    void delete_this(mem_pool* mp) const override { delete_ptr(mp, this); }
};

IncFileDBImpl::IncFileDBImpl(mem_pool* mp, const vector<sh_text>* incDirs) :
    mp_(mp), keyInc_(AllocInt(mp)), keyFull_(AllocInt(mp)), tblData_(AllocData(mp))
{
    try {
        vifp_.reserve(100);

        vector<IncFilePath> stk;
        stk.reserve(10);

        for (auto it=incDirs->begin(), end=incDirs->end(); it!=end; ++it)
            stk.emplace_back(*it, shtx(mp, ""));

        while (!stk.empty()) {
            IncFilePath cur=stk.back();
            stk.pop_back();

            auto names=ex_list_dir(mp, cur.fullPath->str());
            for (auto it=names.begin(), end=names.end(); it!=end; ++it) {
                strsz name=(*it)->str();
                if (name=="." || name=="..") continue;

                IncFilePath next(cur.dirPath, sh_text(path_add(mp, cur.incPath->str(), name)));
                file_info fi=ex_get_info(mp, next.fullPath->str());

                if (fi.type==ft_dir) stk.push_back(next);
                else if (fi.type==ft_file) {
                    vifp_.push_back(next);
                    int pos=vifp_.size()-1;
//fd_file(mp, fd_out).ex_write(tx_buf(mp)+vifp_[pos].to_text()+'\n');

                    {
                        auto key=vifp_[pos].incPath;
                        auto emplaced=keyInc_.emplace(key, pos);
                        if (!emplaced.second) {
                            throw newExc<Exception>(mp, mp, FLLN, tx_buf(mp)+"Duplicated include name: "+key);
                        }
                    }

                    {
                        auto key=vifp_[pos].fullPath;
                        auto emplaced=keyFull_.emplace(key, pos);
                        if (!emplaced.second) {
                            throw newExc<Exception>(mp, mp, FLLN, tx_buf(mp)+"Duplicated include file: "+key);
                        }
                    }
                }
            }
        }
    }
    catch (...) { throw newExc<Exception>(mp, FLLN, "Can't search include directory", recatch(mp, FLLN)); }
}

const IncFilePath* IncFileDBImpl::findByIncName(strsz in) const
{
    auto found=keyInc_.find(shtx(mp_, in));
    if (found==keyInc_.end()) return 0;

    int pos=found->second;
    assert(pos>=0 && pos<int(vifp_.size()));

    return &vifp_[pos];
}

const IncFilePath* IncFileDBImpl::findByFullPath(strsz fp) const
{
    auto found=keyFull_.find(shtx(mp_, fp));
    if (found==keyFull_.end()) return 0;

    int pos=found->second;
    assert(pos>=0 && pos<int(vifp_.size()));

    return &vifp_[pos];
}

const IncFileData* IncFileDBImpl::getData(strsz path)
{
    auto key=shtx(mp_, path);

    auto found=tblData_.find(key);
    if (found!=tblData_.end()) return get(found->second);

    auto data=make_sh<IncFileData>(mp_);

    auto vil=readIncLines(mp_, path);
    for (auto it=vil.begin(), end=vil.end(); it!=end; ++it) {
        sh_text qtdName(mp_);
        if (it->quot) qtdName=(tx_buf(mp_)+'"'+it->name+'"').release();
        else qtdName=(tx_buf(mp_)+'<'+it->name+'>').release();

        data->dirvs.emplace_back(it->line, qtdName, sh_text(resolveIncLine(mp_, path, &(*it), this)));
    }

    tblData_.emplace(key, data);
    return get(data);
}

using on_line_t=void(strsz path, int line, strsz buf, void* arg);

void filterFile(mem_pool* mp, strsz path, on_line_t on_line, void* arg)
{
    fd_file f(mp, path, md_rdo);
    buf_rd br(mp, &f);

    int line=1;
    text buf(mp, 1024);

    bool lineComm=false;
    bool fullComm=false;
    text bufComm(mp, 1024);

    for (;;) {
        int ch;
        while ((ch=br.get())=='\r')
            ;

        if (ch==-1) {
            if (buf.size()) on_line(path, line, buf.str(), arg);
            break;
        }

        if (ch=='\n') {
            on_line(path, line, buf.str(), arg);

            line++;
            buf.clear();
            lineComm=false;
        }
        else {
            if (!lineComm && !fullComm) buf+=ch;
            else if (fullComm) bufComm+=ch;
        }

        if (!fullComm) {
            if (buf.str().ends("//")) {
                buf.uninitialized_resize(buf.size()-2);
                lineComm=true;
            }
            else if (buf.str().ends("/*")) {
                buf.uninitialized_resize(buf.size()-2);

                fullComm=true;
                bufComm.clear();
            }
        }
        else {
            if (bufComm.str().ends("*/")) fullComm=false;
        }
    }
}

struct LineData {
    mem_pool* mp;
    vector<IncLine> vil;

    LineData(mem_pool* m) : mp(m) { vil.reserve(10); }
};

int skipWs(strsz s, int pos)
{
    for ( ; pos<s.size(); pos++) {
        if (!isspace(s[pos])) return pos;
    }

    return pos;
}

void onLine(strsz path, int line, strsz buf, void* arg)
{
    auto ld=static_cast<LineData*>(arg);
    auto mp=ld->mp;
//fd_file(mp, fd_out).ex_write(tx_buf(mp)+line+":"+buf+"\n");

    int p=skipWs(buf, 0);
    if (p>=buf.size() || buf[p]!='#') return;

    p=skipWs(buf, p+1);
    if (p>=buf.size() || buf[p]!='i') return;

    text word(mp, 10);
    for ( ; p<buf.size(); p++) {
        if (!isalpha(buf[p])) break;
        word+=buf[p];
    }
    if (word.str()!="include") return;

    IncLine il(line, shtx(mp, ""), false);
    auto exc=newExc<Exception>(mp, mp, FLLN, tx_buf(mp)+"Invalid #include directive in "+path+":"+line+": "+buf);

    p=skipWs(buf, p);
    if (p>=buf.size()) throw exc;

    if (buf[p]=='<') il.quot=false;
    else if (buf[p]=='"') il.quot=true;
    else throw exc;

    int beg=++p;
    for (;; p++) {
        if (p>=buf.size()) throw exc;
        if ((buf[p]=='"' && il.quot) || (buf[p]=='>' && !il.quot)) {
            *il.name=strsz(buf.str()+beg, p-beg);
            break;
        }
    }

    if (!il.name->size()) throw exc;

    p++;
    for ( ; p<buf.size(); p++) {
        if (!isspace(buf[p])) throw exc;
    }
//fd_file(mp, fd_out).ex_write(tx_buf(mp)+il.to_text()+'\n');

    ld->vil.push_back(il);
}

}  // unnamed

namespace inchk {

IncFilePath::IncFilePath(sh_text dirp, sh_text incp) : dirPath(dirp), incPath(incp), fullPath(pool(dirp))
{
    auto fp=path_add(pool(dirp), dirp->str(), incp->str());

    char sp=last_sepr(dirp->str());
    for (int i=dirp->size(); i<fp->size(); i++) {
        if (is_sepr((*fp)[i])) (*fp)[i]=sp;
    }

    fullPath=move(fp);
}

un_text IncFilePath::to_text() const
{
    return (tx_buf(pool(dirPath))+"dirPath("+dirPath+") incPath("+incPath+") fullPath("+fullPath+")").release();
}

un_text IncDirective::to_text() const
{
    return (tx_buf(pool(qtdName))+"line("+line+") qtdName("+qtdName+") fullPath("+fullPath+")").release();
}

un_ptr<IncFileDB> newIncFileDB(mem_pool* mp, const vector<sh_text>* incDirs)
{
    return un_ptr<IncFileDB>(mp, mp_new<IncFileDBImpl>(mp, mp, incDirs));
}

un_text IncLine::to_text() const
{
    return (tx_buf(pool(name))+"line("+line+") name("+name+") quot("+quot+")").release();
}

vector<IncLine> readIncLines(mem_pool* mp, strsz path)
{
    LineData ld(mp);
    filterFile(mp, path, onLine, &ld);

    return ld.vil;
}

un_text resolveIncLine(mem_pool* mp, strsz path, const IncLine* il, const IncFileDB* db)
{
    if (il->quot) {
        strsz ln=last_name(path);
        assert(ln.size());

        auto incp=untx(mp, strsz(path.str(), path.size()-ln.size()));
        *incp+=il->name->str();

        error_code ec;
        get_info(mp, incp->str(), &ec);

        if (!ec) return incp;
    }

    auto incf=db->findByIncName(il->name->str());
    if (incf) return untx(mp, incf->fullPath->str());

    return untx(mp, "");
}

}  // namespace inchk
