/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#include "print.hpp"

using namespace std;
using namespace ders;
using namespace inchk;

namespace inchk {

struct Args {
    vector<sh_text> incDirs;
    vector<sh_text> srcFiles;
};

un_ptr<Args> parse(mem_pool* mp, int argc, char** argv)
{
    strsz usage("inchk -Iinclude/dir [...] source/file.cpp [...]");
    if (argc<2) throw newExc<ExitMsgException>(mp, mp, FLLN, usage, 1);

    auto ret=make_un<Args>(mp);
    for (char** a=argv+1; *a; a++) {
        strsz arg(*a);
        if (!arg.size()) throw newExc<ExitMsgException>(mp, mp, FLLN, usage, 1);

        if (arg.front()=='-') {
            if (arg.begins("-I")) {
                strsz id(arg.str()+2, arg.size()-2);
                if (!id.size()) throw newExc<ExitMsgException>(mp, mp, FLLN, usage, 1);

                ret->incDirs.push_back(shtx(mp, id));
            }
            else throw newExc<ExitMsgException>(mp, mp, FLLN, usage, 1);
        }
        else ret->srcFiles.push_back(shtx(mp, arg));
    }

    return ret;
}

}  // namespace inchk

int main(int argc, char** argv)
{
    mem_pool mpo, *mp=&mpo;
    fd_file out(mp, fd_out), err(mp, fd_err);

    try {
        auto args=parse(mp, argc, argv);

        auto db=newIncFileDB(mp, &args->incDirs);

        for (auto it=args->srcFiles.begin(), end=args->srcFiles.end(); it!=end; ++it)
            print(&out, (*it)->str(), get(db));

        return 0;
    }
    catch (...) {
        error_code ignore;

        auto se=recatch(mp, FLLN, false);
        if (auto em=se->is<ExitMsgException>()) {
            if (em->message().size()) (em->exitCode() ? err : out).write(tx_buf(mp)+em->message()+'\n', &ignore);
            return em->exitCode();
        }

        err.write(toTextAll(get(se))->str(), &ignore);
        return 2;
    }
}
