/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#include <chrono>
#include <list>
#include <thread>
#include <vector>
#include <ders/fd_file.hpp>
#include <ders/latch.hpp>
#include <ders/mp_allocator.hpp>

using namespace std;
using namespace std::chrono;
using namespace ders;

const int N=5000;
const int M=10000;

void start_std(latch* l1, latch* l2)
{
    list<int> lst;

    l1->arrive_and_wait();
    for (int i=0; i<N; i++) {
        for (int j=0; j<M; j++) lst.push_back(j);
        for (int j=0; j<M; j++) lst.pop_front();
    }
    l2->arrive_and_wait();
}

void start_ders(latch* l1, latch* l2)
{
    mem_pool mpo;
    mp_allocator<int> mpa(&mpo);
    list<int, mp_allocator<int>> lst(mpa);

    l1->arrive_and_wait();
    for (int i=0; i<N; i++) {
        for (int j=0; j<M; j++) lst.push_back(j);
        for (int j=0; j<M; j++) lst.pop_front();
    }
    l2->arrive_and_wait();
}

void steady_now(void* arg)
{
    auto pt=static_cast<steady_clock::time_point*>(arg);
    *pt=steady_clock::now();
}

int main(int argc, char** argv)
{
    mem_pool mpo, *mp=&mpo;
    fd_file out(mp, fd_out), err(mp, fd_err);

    try { 
        if (argc!=3) {
        m1:            
            err.ex_write("main num_threads std|ders");
            return 1;
        }

        int numThr=atoi(argv[1]);
        if ( !(numThr>=1 && numThr<=100) ) {
            err.ex_write("num_threads must be in [1, 100]");
            return 1;
        }

        void (*start)(latch*, latch*);
        if (strsz(argv[2])=="std") start=start_std;
        else if (strsz(argv[2])=="ders") start=start_ders;
        else goto m1;

        steady_clock::time_point t1, t2;
        auto l1=new_latch(mp, numThr, steady_now, &t1);
        auto l2=new_latch(mp, numThr, steady_now, &t2);

        vector<thread> vthr;
        for (int i=0; i<numThr; i++) {
            thread t(start, get(l1), get(l2));
            vthr.push_back(move(t));
        }

        for (int i=0; i<numThr; i++) vthr[i].join();

        out.ex_write(tx_buf(mp)+numThr+'\t'+duration_cast<milliseconds>(t2-t1).count()+'\t'+argv[2]+'\n');
        return 0;
    }
    catch (...) { 
        err.ex_write(toTextAll(get(recatch(mp, FLLN, false)))->str());
        return 2;
    }
}
