/*
 * Copyright (c) Sergey P. Derevyago, 2023.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * http://www.boost.org/LICENSE_1_0.txt
 *
 */

#include <chrono>
#include <fstream>
#include <memory>
#include <thread>
#include <vector>
#include <ders/fd_file.hpp>
#include <ders/latch.hpp>

using namespace std;
using namespace std::chrono;
using namespace ders;

struct ThreadData {
    thread thr;
    latch* lch1;
    latch* lch2;

    const char* path;

    const char* err=0;
    int sum=0;

    ThreadData(latch* l1, latch* l2, const char* p) : lch1(l1), lch2(l2), path(p) {}
};

void start_c(ThreadData* td)
{
    FILE* f=fopen(td->path, "rb");
    if (!f) {
        td->err="can't open file";
        return;
    }

    td->lch1->arrive_and_wait();
    for (int ch; (ch=getc(f))!=EOF; )
        td->sum+=ch;
    td->lch2->arrive_and_wait();

    fclose(f);
}

void start_cpp(ThreadData* td)
{
    ifstream f(td->path, ios_base::in|ios_base::binary);
    if (!f) {
        td->err="can't open file";
        return;
    }

    td->lch1->arrive_and_wait();
    for (int ch; (ch=f.get())!=EOF; )
        td->sum+=ch;
    td->lch2->arrive_and_wait();
}

void start_ders(ThreadData* td)
{
    mem_pool mpo, *mp=&mpo;
    
    error_code ec;
    fd_file f(mp, strsz(td->path), md_rdo, &ec);
    if (ec) {
        td->err="can't open file";
        return;
    }

    buf_rd bf(mp, &f);

    td->lch1->arrive_and_wait();
    for (int ch; (ch=bf.get())!=-1; )
        td->sum+=ch;
    td->lch2->arrive_and_wait();
}

void steady_now(void* arg)
{
    auto pt=static_cast<steady_clock::time_point*>(arg);
    *pt=steady_clock::now();
}

int main(int argc, char** argv)
{
    mem_pool mpo, *mp=&mpo;
    fd_file out(mp, fd_out), err(mp, fd_err);

    try { 
        if (argc!=4) {
        m1:            
            err.ex_write("main file num_threads c|cpp|ders");
            return 1;
        }

        int numThr=atoi(argv[2]);
        if ( !(numThr>=1 && numThr<=100) ) {
            err.ex_write("num_threads must be in [1, 100]");
            return 1;
        }

        strsz a3(argv[3]);
        void (*start)(ThreadData* td);

        if (a3=="c") start=start_c;
        else if (a3=="cpp") start=start_cpp;
        else if (a3=="ders") start=start_ders;
        else goto m1;

        steady_clock::time_point t1, t2;
        auto l1=new_latch(mp, numThr, steady_now, &t1);
        auto l2=new_latch(mp, numThr, steady_now, &t2);

        vector<ThreadData> vtd;
        vtd.reserve(numThr);

        for (int i=0; i<numThr; i++) {
            vtd.emplace_back(get(l1), get(l2), argv[1]);

            thread thr(start, &vtd.back());
            vtd.back().thr=move(thr);
        }

        for (int i=0; i<numThr; i++) vtd[i].thr.join();

        for (int i=0; i<numThr; i++) {
            if (vtd[i].err) {
                err.ex_write(tx_buf(mp)+"thread "+i+": "+vtd[i].err);
                return 2;
            }

            if (vtd[i].sum!=vtd[0].sum) {
                err.ex_write(tx_buf(mp)+"thread "+i+": bad file sum");
                return 2;
            }
        }

        out.ex_write(tx_buf(mp)+numThr+'\t'+duration_cast<milliseconds>(t2-t1).count()+'\t'+a3+"\tsum="+vtd[0].sum+'\n');
        return 0;
    }
    catch (...) { 
        err.ex_write(toTextAll(get(recatch(mp, FLLN, false)))->str());
        return 2;
    }
}
